"""
    item 流程控制器
"""

from loguru import logger
from palp import settings
from threading import Thread
from palp.item.item import Item
from palp.exception import DropItemException
from palp.tool.short_module import import_module, sort_module


class ItemController(Thread):
    PIPELINE = []

    def __new__(cls, *args, **kwargs):
        """
        实例化之前加载设置

        :param args:
        :param kwargs:
        """
        if not cls.PIPELINE:
            cls.from_settings()

        return object.__new__(cls)

    def __init__(self, spider, q):
        super().__init__()
        self.spider = spider
        self.queue = q

        self.item_buffer = []
        self.item_buffer_max_size = settings.PIPELINE_ITEM_BUFFER  # item 最大存储数量

    @classmethod
    def from_settings(cls):
        """
        导入设置

        :return:
        """
        modules = sort_module(
            cls_dict=settings.PIPELINE,
            palp_cls_dict=settings.PALP_PIPELINE,
            cls_mapping={
                'ITEM_FILTER_PIPELINE': settings.ITEM_FILTER_PIPELINE[settings.SPIDER_TYPE][settings.FILTERING_MODE]
            }
        )

        cls.PIPELINE = import_module(modules)

    def run(self) -> None:
        """
        item 处理入口

        :return:
        """
        try:
            while True:
                try:
                    if self.spider.spider_done and self.queue.empty():
                        if self.item_buffer:
                            self.pipeline_save()
                        break

                    task = self.queue.get(timeout=1)
                    if task is None:
                        continue

                    # 每个 item 都会先走处理
                    self.pipeline_in(item=task)

                    # 判断是否需要入库
                    if self.item_buffer_max_size == 0 or self.buffer_size >= self.item_buffer_max_size:
                        self.pipeline_save()

                except DropItemException:
                    logger.debug(f"丢弃 item：{self.item_buffer}")
                except Exception as e:
                    # spider 报错处理
                    for middleware in self.spider.SPIDER_MIDDLEWARE:
                        middleware.spider_error(self.spider, e)
        finally:
            self.pipeline_close()

    def pipeline_in(self, item: Item):
        """
        处理清洗等

        :param item:
        :return:
        """
        for pipeline in self.__class__.PIPELINE:
            pipeline.pipeline_in(self.spider, item)

        self.item_buffer.append(item)

    def pipeline_save(self):
        """
        处理入库

        :return:
        """
        # 禁用的话就单条
        if self.item_buffer_max_size == 0 and len(self.item_buffer) != 0:
            item_buffer = self.item_buffer[0]
        else:
            item_buffer = self.item_buffer

        # 入库
        try:
            failed_times = 0
            while failed_times < settings.PIPELINE_RETRY_TIMES:
                try:
                    for pipeline in self.__class__.PIPELINE:
                        pipeline.pipeline_save(self.spider, item_buffer)
                    break
                except DropItemException:
                    raise
                except Exception as e:
                    failed_times += 1

                    for pipeline in self.__class__.PIPELINE:
                        pipeline.pipeline_error(self.spider, item_buffer, e)

                if failed_times >= settings.PIPELINE_RETRY_TIMES:
                    for pipeline in self.__class__.PIPELINE:
                        pipeline.pipeline_failed(self.spider, item_buffer)
                    return
        finally:
            self.item_buffer.clear()

    def pipeline_close(self):
        """
        spider 结束时执行

        :return:
        """
        for pipeline in self.__class__.PIPELINE:
            pipeline.pipeline_close(self.spider)

    @property
    def buffer_size(self):
        """
        buffer 大小

        :return:
        """
        return len(self.item_buffer)



"""
    item 去重

    注意：
        setting 启用才会去重
"""
from palp import settings
from palp.filter.filter_bloom import BloomFilter
from palp.pipeline.pipeline import Pipeline
from palp.filter.filter_redis_set import RedisSetFilter
from palp.filter.filter_set import SetFilter
from palp.exception import DropItemException
from palp.filter.filter_redis_bloom import RedisBloomFilter


class SetFilterPipeline(Pipeline):
    """
        基于内存的 set 的去重
    """

    def __init__(self):
        self.item_memory_filter = SetFilter()

    def pipeline_in(self, spider, item):
        if settings.FILTER_ITEM:
            is_repeat = self.item_memory_filter.is_repeat(spider=spider, obj=item)

            if is_repeat:
                raise DropItemException(f"丢弃重复 item：{item}")


class BloomFilterPipeline(Pipeline):
    """
        基于内存的 bloom 的去重
    """

    def __init__(self):
        self.item_bloom_filter = BloomFilter()

    def pipeline_in(self, spider, item):
        if settings.FILTER_ITEM:
            is_repeat = self.item_bloom_filter.is_repeat(spider=spider, obj=item)

            if is_repeat:
                raise DropItemException(f"丢弃重复 item：{item}")


class RedisSetFilterPipeline(Pipeline):
    """
        基于 redis 的 set 的去重
    """

    def __init__(self):
        self.item_redis_filter = RedisSetFilter()

    def pipeline_in(self, spider, item):
        if settings.FILTER_ITEM:
            is_repeat = self.item_redis_filter.is_repeat(spider=spider, obj=item)

            if is_repeat:
                raise DropItemException(f"丢弃重复 item：{item}")


class RedisBloomFilterPipeline(Pipeline):
    """
        基于 redis 的 bloom 去重
    """

    def __init__(self):
        self.item_redis_bloom_filter = RedisBloomFilter()

    def pipeline_in(self, spider, item):
        if settings.FILTER_ITEM:
            is_repeat = self.item_redis_bloom_filter.is_repeat(spider=spider, obj=item)

            if is_repeat:
                raise DropItemException(f"丢弃重复 item：{item}")



"""
    通过 redis 实现心跳功能
"""
import json
import time
import uuid
import datetime
import threading
from loguru import logger
from palp import settings
from quickdb import RedisLockNoWait, RedisLock


class ClientHeart:
    def __init__(self, spider):
        """

        :param spider:
        """
        self.spider = spider

        self.beating_time = 4  # 心跳频率
        self.check_time = self.beating_time - 1  # 心跳检查频率
        self.client_name = self.generate_client_name()  # 随机客户端名

    def start(self):
        beating = threading.Thread(target=self.beating, daemon=True)
        beating.start()

        try:
            self.check_client_beating()
        finally:
            self.close(beating)

    def check_client_beating(self):
        """
        检查各个客户端的心跳，超过 2 次视为停止

        注意：因为不是需要每个都运行，而是同时有一个去运行，所以使用了 RedisLockNoWait
        :return:
        """
        from palp.conn import redis_conn

        while not self.all_client_is_waiting:

            all_client_is_waiting = True

            with RedisLockNoWait(conn=redis_conn, lock_name=settings.REDIS_KEY_LOCK, block_timeout=20) as lock:
                # 判断是否上锁成功
                if not lock.lock_success:
                    continue

                # 上锁成功则判断客户端运行情况
                for client_name, detail in redis_conn.hgetall(settings.REDIS_KEY_HEARTBEAT).items():
                    client_name = client_name.decode()
                    detail = json.loads(detail.decode())

                    # 校验 2 次失败则为客户端关闭
                    if int(time.time()) - detail['time'] - self.beating_time > 1:
                        if redis_conn.sismember(settings.REDIS_KEY_HEARTBEAT_FAILED, client_name):
                            logger.warning(f"该客户端异常关闭：{client_name}")
                            redis_conn.srem(settings.REDIS_KEY_HEARTBEAT_FAILED, client_name)
                            redis_conn.hdel(settings.REDIS_KEY_HEARTBEAT, client_name)
                        else:
                            logger.warning(f"该客户端可能异常关闭：{client_name}")
                            redis_conn.sadd(settings.REDIS_KEY_HEARTBEAT_FAILED, client_name)

                            # 检查是否是 master 死机，是的话自己成为 master
                            master_name = redis_conn.get(settings.REDIS_KEY_MASTER)
                            if master_name and master_name.decode() == client_name:
                                self.spider.spider_master = True
                                redis_conn.set(settings.REDIS_KEY_MASTER, self.client_name)
                    else:
                        if redis_conn.sismember(settings.REDIS_KEY_HEARTBEAT_FAILED, client_name):
                            redis_conn.srem(settings.REDIS_KEY_HEARTBEAT_FAILED, client_name)

                        if detail['waiting'] is False:
                            all_client_is_waiting = False
                        logger.debug(f"心跳正常：{client_name}")

                # 如果所有客户端都无任务，则结束
                if all_client_is_waiting:
                    logger.debug("所有客户端都已挂起，即将停止")
                    self.stop_all_client()
                    break

                time.sleep(self.check_time)

            time.sleep(1)  # 避免访问频繁

    def beating(self):
        """
        保持心跳

        :return:
        """
        from palp.conn import redis_conn

        while not self.all_client_is_waiting:
            redis_conn.hset(
                settings.REDIS_KEY_HEARTBEAT,
                self.client_name,
                json.dumps({
                    "time": int(time.time()),
                    "waiting": self.spider.all_spider_controller_is_waiting(),
                }, ensure_ascii=False)
            )
            time.sleep(self.beating_time)

    def close(self, beating: threading.Thread):
        """
        清理资源

        :return:
        """
        from palp.conn import redis_conn

        while True:
            if not beating.is_alive():
                redis_conn.hdel(settings.REDIS_KEY_HEARTBEAT, self.client_name)
                break

    @staticmethod
    def stop_all_client():
        """
        发出停止信号，让所有程序停止

        :return:
        """
        from palp.conn import redis_conn

        redis_conn.set(settings.REDIS_KEY_STOP, '停止时间：' + str(datetime.datetime.now()))

    @property
    def all_client_is_waiting(self) -> bool:
        """
        判断是否所有客户端都陷入了等待

        :return:
        """
        from palp.conn import redis_conn

        return bool(redis_conn.exists(settings.REDIS_KEY_STOP))

    @staticmethod
    def remove_stop_status():
        """
        移除停止信号

        :return:
        """
        from palp.conn import redis_conn

        while True:
            if not redis_conn.exists(settings.REDIS_KEY_HEARTBEAT):
                redis_conn.delete(settings.REDIS_KEY_STOP)
                break

            time.sleep(0.1)

    def generate_client_name(self) -> str:
        """
        生成客户端的名字

        :return:
        """
        from palp.conn import redis_conn

        while True:
            name = str(uuid.uuid1())
            with RedisLock(conn=redis_conn, lock_name=settings.REDIS_KEY_LOCK, block_timeout=10):
                if redis_conn.hget(settings.REDIS_KEY_HEARTBEAT, name):
                    continue
                else:

                    if self.spider.spider_master:
                        redis_conn.set(settings.REDIS_KEY_MASTER, name)

                    redis_conn.hset(settings.REDIS_KEY_HEARTBEAT, name, json.dumps({
                        "time": int(time.time()),
                        "waiting": False,
                    }, ensure_ascii=False))
                    return name



"""
    请求的处理
"""
import time
import random
import urllib3
from palp import settings
from typing import Callable
from urllib.parse import urlparse
from palp.network.response import Response
from palp.tool.user_agent import random_ua
from requests.cookies import RequestsCookieJar
from palp.tool.short_module import import_module
from urllib3.exceptions import InsecureRequestWarning

# 禁用不安全警告
urllib3.disable_warnings(InsecureRequestWarning)


class Request:
    # requests 模块所需的
    __REQUEST_ATTRS__ = [
        'url',
        'method',
        'params',
        'data',
        'headers',
        'cookies',
        'files',
        'auth',
        'timeout',
        'allow_redirects',
        'proxies',
        'hooks',
        'stream',
        'verify',
        'cert',
        'json',
    ]

    # 框架运行所需的
    __PALP_ATTRS__ = [
        'filter_repeat',
        'keep_session',
        'keep_cookie',
        'new_session',
        'callback',
        'cookie_jar',
        'priority',
        'command'
    ]

    # 下载器
    DOWNLOADER = None
    DOWNLOADER_PARSER = None

    def __new__(cls, *args, **kwargs):
        """
        导入下载器

        :param args:
        :param kwargs:
        """
        if cls.DOWNLOADER is None:
            cls.DOWNLOADER = import_module(settings.RESPONSE_DOWNLOADER, instantiate=False)[0]
            cls.DOWNLOADER_PARSER = import_module(settings.RESPONSE_DOWNLOADER_PARSER, instantiate=False)[0]

        return object.__new__(cls)

    def __init__(
            self,
            url: str,
            method: str = None,
            params=None,
            data=None,
            headers=None,
            cookies=None,
            timeout=None,
            proxies=None,
            json=None,
            downloader=None,
            filter_repeat: bool = False,
            keep_session: bool = False,
            keep_cookie: bool = False,
            callback: Callable = None,
            cookie_jar: RequestsCookieJar = None,
            priority: int = settings.DEFAULT_QUEUE_PRIORITY,
            command: dict = None,
            **kwargs
    ):
        """
        requests 参数
        :param url: 请求链接
        :param params: params
        :param data: data
        :param headers: 请求头
        :param cookies: cookies
        :param timeout: 超时时间
        :param proxies: 代理：{'http':'http://ip:port','https':'https://ip:port'}
        :param json: json：也可以使用 data = json.dumps(data)

        Palp 参数
        :param filter_repeat: 是否过滤请求，settings.FILTER_REQUEST 启用生效，默认 False
        :param keep_session: 是否保持 session，默认 False
        :param keep_cookie: 不使用 session 时保持 cookie，默认 False
        :param callback: 回调函数
        :param priority: 启用优先级队列时的优先级，分数越大，优先级约低，默认 settings.DEFAULT_QUEUE_PRIORITY
        :param command: 自定义操作命令，用于自定义 downloader 时使用

        Palp 参数（非用户设置）
        :param cookie_jar: cookie_jar，存储 cookie，这里使用的是 requests 模块的，其它请求的话可以自己提取

        传递的参数
        :param kwargs: 其余需要传递的参数，若参数名不存在则返回 None
        """

        # Request 所需字段
        self._requests_params = {}  # request 参数
        self.command = command
        self.callback = callback
        self.priority = priority
        self.downloader = downloader
        self.cookie_jar = cookie_jar
        self.keep_cookie = keep_cookie
        self.keep_session = keep_session
        self.filter_repeat = filter_repeat

        # requests 请求参数
        self.url = url
        self.data = data
        self.json = json
        self.method = method
        self.params = params
        self.headers = headers
        self.cookies = cookies
        self.timeout = timeout
        self.proxies = proxies

        # 传递参数更新一下
        for key, value in kwargs.items():
            self[key] = value

    def send(self) -> Response:
        """
        获取响应

        :return: ResponseParser 的解析器
        """
        # 加载默认值
        self.set_default()

        # 判断是否等待
        if isinstance(settings.REQUEST_DELAY, int) or isinstance(settings.REQUEST_DELAY, float):
            time.sleep(settings.REQUEST_DELAY)
        elif isinstance(settings.REQUEST_DELAY, list):
            time.sleep(random.choice([i for i in range(settings.REQUEST_DELAY[0], settings.REQUEST_DELAY[1] + 1)]))

        # 获取响应
        response = self.downloader(
            **self._requests_params,
            keep_session=self.keep_session,
            keep_cookie=self.keep_cookie,
            cookie_jar=self.cookie_jar,
            command=self.command
        ).response()

        response_parser = self.__class__.DOWNLOADER_PARSER(response)  # 解析器解析响应
        if response_parser.cookies:
            self.cookie_jar.update(response_parser.cookies)  # 将响应的 cookie 更新到 cookieJar

        return response_parser

    def set_default(self):
        """
        设置一些默认值

        :return:
        """
        # 判断类型
        if self.method:
            self.method = self.method.upper()
        elif self.data or self.json:
            self.method = 'POST'
        else:
            self.method = 'GET'

        # 设置默认
        if self.downloader is None:
            self.downloader = self.__class__.DOWNLOADER
        if self.cookie_jar is None:
            self.cookie_jar = RequestsCookieJar()
        if self.command is None:
            self.command = {}
        if self.headers is None:
            self.headers = {}
        if self.cookies is None:
            self.cookies = {}
        if self.timeout is None:
            self.timeout = settings.REQUEST_TIMEOUT or 60
        self._requests_params.setdefault('verify', False)

        # 添加 ua（模块在使用 ua 的时候可能访问降速）
        if settings.RANDOM_USERAGENT:
            ua = random_ua()
        else:
            ua = settings.DEFAULT_USER_AGENT

        if not self.headers:
            self.headers.update({'User-Agent': ua})
        elif self.headers.get('User-Agent'):
            self.headers.update({'User-Agent': ua})
        elif self.headers.get('user-agent'):
            self.headers.update({'user-agent': ua})

    def to_dict(self) -> dict:
        """
        获取字典形式

        主要作用：快速进行二次请求

        示例：
            request_dict = request.to_dict()
            request_dict[xxx] = xxx # 修改

            yield palp.Request(**request_dict)
        :return:
        """
        request_dict = {}

        for key, value in self.__dict__.items():
            if key.startswith('_') or not value:
                continue
            request_dict[key] = value

        return request_dict

    @property
    def domain(self) -> str:
        """
        获取请求的域名

        :return:
        """
        return urlparse(self.url).netloc

    def __getattr__(self, item):
        """
        可以 request.xxx 进行访问，但是这里严格一点没有就报错避免误导

        :param item:
        :return:
        """
        if item in self.__dict__:
            return self.__dict__[item]

        raise AttributeError(f'未定义的属性：{item}')

    def __setattr__(self, key, value):
        """
        实现 requests.xxx 设置参数（只有 self.xxx 才会进入该函数）

        @param key:
        @param value:
        @return:
        """
        self.__dict__[key] = value

        if key in self.__class__.__REQUEST_ATTRS__:
            self._requests_params[key] = value

    def __lt__(self, other):
        """
        返回比较，否则使用优先级队列会报错

        :param other:
        :return:
        """
        return self.priority < other.priority

    def __str__(self):
        return f"<Request {self.method}-{self.url}>"



"""
    分布式 spider
"""
import json
import time
import importlib
from loguru import logger
from palp import settings
from abc import abstractmethod
from quickdb import RedisLockNoWait
from palp.spider.spider import Spider
from palp.network.request import Request
from palp.tool.client_heart import ClientHeart
from requests.cookies import RequestsCookieJar
from palp.tool.short_module import import_module
from palp.sequence.sequence_redis_item import FIFOItemRedisSequence
from palp.decorator.decorator_spider_middleware import SpiderMiddlewareDecorator


class DistributiveSpider(Spider):
    """
        分布式 spider
    """

    def __new__(cls, *args, **kwargs):
        """
        加载 spider 设置，并定义 SPIDER_TYPE

        :param args:
        :param kwargs:
        """
        if not cls.SPIDER_MIDDLEWARE:
            setattr(settings, 'SPIDER_TYPE', 2)
            cls.from_settings()

        return object.__new__(cls)

    def __init__(self, thread_count=None, redis_key: str = None, request_filter=False, item_filter=False):
        """

        :param thread_count: 线程数量
        :param redis_key: 自定义 redis_key 的名字
        :param request_filter: 开启请求过滤，不然 filter_repeat 无效
        :param item_filter: 开启 item 过滤
        """
        self.spider_master = False
        self.redis_key = redis_key or self.name

        # 根据 spider 的名字设置 redis 前缀
        for key, value in settings.__dict__.items():
            if key.startswith('REDIS_KEY'):
                setattr(settings, key, value.format(redis_key=self.redis_key))

        super().__init__(thread_count, request_filter, item_filter)

        queue_module = settings.REQUEST_QUEUE[settings.SPIDER_TYPE][settings.REQUEST_QUEUE_MODE]
        self.queue = import_module(queue_module)[0]  # 请求队列
        self.queue_item = FIFOItemRedisSequence()  # item 队列

    @abstractmethod
    def start_requests(self) -> None:
        """
        起始函数

        :return:
        """
        pass

    @staticmethod
    def start_check():
        """
        启动检查，防止上次意外结束，导致 master 死机 key 未删除，无法正常启动
        这里设定时间 30s 即出意外后，超过 30s 启动才会移除 master key

        :return:
        """
        from palp.conn import redis_conn

        master_name = redis_conn.get(settings.REDIS_KEY_MASTER)
        if master_name == b'':
            redis_conn.delete(settings.REDIS_KEY_MASTER)
        elif master_name:
            master_detail = redis_conn.hget(settings.REDIS_KEY_HEARTBEAT, master_name.decode())
            if master_detail and time.time() - json.loads(master_detail.decode())['time'] > 30:
                redis_conn.delete(settings.REDIS_KEY_MASTER)

    def start_distribute_failed_request(self):
        """
        重新放入 执行失败的请求任务

        :return:
        """
        from palp.conn import redis_conn

        if settings.REQUEST_RETRY_FAILED and redis_conn.exists(settings.REDIS_KEY_QUEUE_BAD_REQUEST):
            while True:
                request = redis_conn.spop(settings.REDIS_KEY_QUEUE_BAD_REQUEST, count=1)
                if not request:
                    break

                request = json.loads(request[0].decode())

                # 起始函数无 callback 默认添加
                if request['callback'] is None:
                    request['callback'] = 'parse'
                elif isinstance(request['callback'], str) and hasattr(self, request['callback']):
                    pass
                else:
                    logger.warning(f"callback 不是 {self.name} 含有的函数：{request}")
                    continue

                request = Request(**request)

                # 为每一个起始函数添加一个 session、cookie_jar
                request.cookie_jar = RequestsCookieJar()
                request.cookie_jar.update(request['cookies'])
                del request['cookies']

                self.queue.put(request)

    def start_distribute_failed_item(self) -> None:
        """
        重新放入 处理失败的 item

        :return:
        """
        from palp.conn import redis_conn

        modules = {}
        if settings.ITEM_RETRY_FAILED and redis_conn.exists(settings.REDIS_KEY_QUEUE_BAD_ITEM):
            while True:
                item = redis_conn.spop(settings.REDIS_KEY_QUEUE_BAD_ITEM, count=1)
                if not item:
                    break

                item = json.loads(item[0].decode())

                # 重新导入 item
                key = item['module'] + '.' + item['init']
                if key not in modules:
                    cls = importlib.import_module(item['module']).__getattribute__(item['init'])  # 导入
                    modules[key] = cls
                else:
                    cls = modules[key]

                self.queue_item.put(cls(**json.loads(item['data'])))

    def competition_for_master(self) -> None:
        """
        竞争为 master

        :return:
        """
        from palp.conn import redis_conn

        if redis_conn.exists(settings.REDIS_KEY_MASTER):
            return

        with RedisLockNoWait(conn=redis_conn, lock_name=settings.REDIS_KEY_LOCK) as lock:
            if lock.lock_success:
                self.spider_master = True
                redis_conn.set(settings.REDIS_KEY_MASTER, '')

    @SpiderMiddlewareDecorator()
    def run(self) -> None:
        """
        分布式处理 逻辑

        :return:
        """
        from palp.conn import redis_conn

        self.start_controller()  # 任务处理
        self.start_check()  # 检查是否正常
        self.competition_for_master()  # 竞争为 master

        # master 机器处理的事情
        if self.spider_master:
            # 删除停止标志
            redis_conn.delete(settings.REDIS_KEY_STOP)
            # 清空心跳
            redis_conn.delete(settings.REDIS_KEY_HEARTBEAT, settings.REDIS_KEY_HEARTBEAT_FAILED)

            # 分发失败的任务
            self.start_distribute_failed_request()
            self.start_distribute_failed_item()

            # 分发任务
            self.start_distribute()
        else:
            time.sleep(1)  # 睡 1s 避免停止标志没被移除

        # 等待所有任务执行结束
        ClientHeart(spider=self).start()

        # master 机器处理的事情
        if self.spider_master:
            # 删除 master 的标志
            redis_conn.delete(settings.REDIS_KEY_MASTER)
            # 删除 stop 标志
            ClientHeart.remove_stop_status()
            # 删除 heart 标志
            redis_conn.delete(settings.REDIS_KEY_HEARTBEAT, settings.REDIS_KEY_HEARTBEAT_FAILED)
            # 判断是否移除 filter
            if not settings.PERSISTENCE_REQUEST_FILTER:
                redis_conn.delete(settings.REDIS_KEY_QUEUE_FILTER_REQUEST)
            if not settings.PERSISTENCE_ITEM_FILTER:
                redis_conn.delete(settings.REDIS_KEY_QUEUE_FILTER_ITEM)



"""
    等待 spider 各项功能执行结束
"""
from palp.middleware.middleware_spider import SpiderMiddleware


class SpiderWaitMiddleware(SpiderMiddleware):
    """
        等待功能执行结束
    """

    def spider_close(self, spider):
        """
        等待 spider controller、item controller 执行完毕
        :param spider:
        :return:
        """
        spider.wait_spider_controller_done()
        spider.wait_item_controller_done()



from palp.pipeline.pipeline import Pipeline
from palp.pipeline.pipeline_recycle import RedisRecyclePipeline
from palp.pipeline.pipeline_filter import SetFilterPipeline, BloomFilterPipeline, RedisSetFilterPipeline, \
    RedisBloomFilterPipeline
